// Copyright (c) Bartłomiej Płotka @bwplotka
// Licensed under the Apache License 2.0.

// Package mod allows programmatic editions of Go mod file. It wraps `golang.org/x/mod` with easier to use API.
package mod
